@@ -1,11 +1,11 @@
 ;;; helpful.el --- a better *help* buffer            -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2017-2018  Wilfred Hughes
+;; Copyright (C) 2017-2019  Wilfred Hughes
 
 ;; Author: Wilfred Hughes <me@wilfred.me.uk>
 ;; URL: https://github.com/Wilfred/helpful
 ;; Keywords: help, lisp
-;; Version: 0.15
+;; Version: 0.16
 ;; Package-Requires: ((emacs "25.1") (dash "2.12.0") (dash-functional "1.2.0") (s "1.11.0") (f "0.20.0") (elisp-refs "1.2") (shut-up "0.3"))
 
 ;; This program is free software; you can redistribute it and/or modify
@@ -60,8 +60,10 @@
 (defvar-local helpful--sym nil)
 (defvar-local helpful--callable-p nil)
 (defvar-local helpful--associated-buffer nil
-  "We store a reference to the buffer we were called from, so we can
-show the value of buffer-local variables.")
+  "The buffer being used when showing inspecting
+buffer-local variables.")
+(defvar-local helpful--start-buffer nil
+  "The buffer we were originally called from.")
 (defvar-local helpful--view-literal nil
   "Whether to show a value as a literal, or a pretty interactive
 view.")
@@ -141,7 +143,11 @@ can make Helpful very slow.")
       (helpful-mode)
       (setq helpful--sym symbol)
       (setq helpful--callable-p callable-p)
-      (setq helpful--associated-buffer current-buffer))
+      (setq helpful--start-buffer current-buffer)
+      (setq helpful--associated-buffer current-buffer)
+      (if (helpful--primitive-p symbol callable-p)
+          (setq-local comment-start "//")
+        (setq-local comment-start ";")))
     buf))
 
 (defface helpful-heading
@@ -152,6 +158,10 @@ can make Helpful very slow.")
   "Propertize TEXT as a heading."
   (format "%s\n" (propertize text 'face 'helpful-heading)))
 
+(defun helpful--warning (text)
+  "Propertize TEXT as a warning."
+  (format "%s\n" (propertize text 'face 'warning)))
+
 (defun helpful--format-closure (sym form)
   "Given a closure, return an equivalent defun form."
   (-let (((_keyword _env args . body) form)
@@ -233,10 +243,12 @@ Return SYM otherwise."
          (push s aliases))))
     (helpful--sort-symbols aliases)))
 
+(defun helpful--obsolete-info (sym callable-p)
+  (when (symbolp sym)
+    (get sym (if callable-p 'byte-obsolete-info 'byte-obsolete-variable))))
+
 (defun helpful--format-alias (sym callable-p)
-  (let ((obsolete-info (if callable-p
-                           (get sym 'byte-obsolete-info)
-                         (get sym 'byte-obsolete-variable)))
+  (let ((obsolete-info (helpful--obsolete-info sym callable-p))
         (sym-button (helpful--button
                      (symbol-name sym)
                      'helpful-describe-exactly-button
@@ -503,12 +515,35 @@ If narrowing is in effect, widen if POS isn't in the narrowed area."
 
 This is largely equivalent to `read-buffer', but counsel.el
 overrides that to include previously opened buffers."
-  (get-buffer
-   (completing-read
-    prompt
-    (-map #'buffer-name (buffer-list))
-    predicate
-    t)))
+  (let* ((names (-map #'buffer-name (buffer-list)))
+         (default
+           (cond
+            ;; If we're already looking at a buffer-local value, start
+            ;; the prompt from the relevant buffer.
+            ((and helpful--associated-buffer
+                  (buffer-live-p helpful--associated-buffer))
+             (buffer-name helpful--associated-buffer))
+            ;; If we're looking at the global value, offer the initial
+            ;; buffer.
+            ((and helpful--start-buffer
+                  (buffer-live-p helpful--start-buffer))
+             (buffer-name helpful--start-buffer))
+            ;; If we're looking at the global value and have no initial
+            ;; buffer, choose the first normal buffer.
+            (t
+             (--first (and (not (s-starts-with-p " " it))
+                           (not (s-starts-with-p "*" it)))
+                      names))
+            )))
+    (get-buffer
+     (completing-read
+      prompt
+      names
+      predicate
+      t
+      nil
+      nil
+      default))))
 
 (defun helpful--associated-buffer (button)
   "Change the associated buffer, so we can see buffer-local values."
@@ -651,10 +686,10 @@ overrides that to include previously opened buffers."
       (insert (format "Functions called by %s:\n\n" sym))
       (helpful--display-callee-group compounds)
 
-      (insert "\n")
-
-      (insert (format "Primitives called by %s:\n\n" sym))
-      (helpful--display-callee-group primitives)
+      (when primitives
+        (insert "\n")
+        (insert (format "Primitives called by %s:\n\n" sym))
+        (helpful--display-callee-group primitives))
 
       (goto-char (point-min))
 
@@ -1174,6 +1209,59 @@ If .elc files exist without the corresponding .el, return nil."
       (find-library-name library-name)
     (error nil)))
 
+(defun helpful--macroexpand-try (form)
+  "Try to fully macroexpand FORM.
+If it fails, attempt to partially macroexpand FORM."
+  (catch 'result
+    (ignore-errors
+      ;; Happy path: we can fully expand the form.
+      (throw 'result (macroexpand-all form)))
+    (ignore-errors
+      ;; Attempt one level of macroexpansion.
+      (throw 'result (macroexpand-1 form)))
+    ;; Fallback: just return the original form.
+    form))
+
+(defun helpful--tree-any-p (pred tree)
+  "Walk TREE, applying PRED to every subtree.
+Return t if PRED ever returns t."
+  (cond
+   ((null tree) nil)
+   ((funcall pred tree) t)
+   ((not (consp tree)) nil)
+   (t (or
+       (helpful--tree-any-p pred (car tree))
+       (helpful--tree-any-p pred (cdr tree))))))
+
+(defun helpful--find-by-macroexpanding (buf sym callable-p)
+  "Search BUF for the definition of SYM by macroexpanding
+interesting forms in BUF."
+  (catch 'found
+    (with-current-buffer buf
+      (save-excursion
+        (goto-char (point-min))
+        (condition-case nil
+            (while t
+              (let ((form (read (current-buffer)))
+                    (var-def-p
+                     (lambda (sexp)
+                       (and (eq (car-safe sexp) 'defvar)
+                            (eq (car-safe (cdr sexp)) sym))))
+                    (fn-def-p
+                     (lambda (sexp)
+                       ;; `defun' ultimately expands to `defalias'.
+                       (and (eq (car-safe sexp) 'defalias)
+                            (equal (car-safe (cdr sexp)) `(quote ,sym))))))
+                (setq form (helpful--macroexpand-try form))
+
+                (when (helpful--tree-any-p
+                       (if callable-p fn-def-p var-def-p)
+                       form)
+                  ;; `read' puts point at the end of the form, so go
+                  ;; back to the start.
+                  (throw 'found (scan-sexps (point) -1)))))
+          (end-of-file nil))))))
+
 (defun helpful--definition (sym callable-p)
   "Return a list (BUF POS OPENED) where SYM is defined.
 
@@ -1239,7 +1327,12 @@ buffer."
             (save-restriction
               (widen)
               (setq pos
-                    (cdr (find-function-search-for-symbol sym nil library-name))))))))
+                    (cdr (find-function-search-for-symbol sym nil library-name))))))
+        ;; If we found the containing buffer, but not the symbol, attempt
+        ;; to find it by macroexpanding interesting forms.
+        (when (and buf (not pos))
+          (setq pos (helpful--find-by-macroexpanding buf sym t)))))
+     ;; A function, but no file found.
      (callable-p
       ;; Functions defined interactively may have an edebug property
       ;; that contains the location of the definition.
@@ -1374,22 +1467,21 @@ alist with the lists concatenated."
                 l1)))
     (append l1-with-values l2-extra-values)))
 
-(defun helpful--keymaps-containing-aliases (command-sym)
+(defun helpful--keymaps-containing-aliases (command-sym aliases)
   "Return a list of pairs mapping keymap symbols to the
 keybindings for COMMAND-SYM in each keymap.
 
 Includes keybindings for aliases, unlike
 `helpful--keymaps-containing'."
-  (let* ((aliases (helpful--aliases command-sym t))
-         (syms (cons command-sym aliases))
+  (let* ((syms (cons command-sym aliases))
          (syms-keymaps (-map #'helpful--keymaps-containing syms)))
     (-reduce #'helpful--merge-alists syms-keymaps)))
 
-(defun helpful--format-keys (command-sym)
+(defun helpful--format-keys (command-sym aliases)
   "Describe all the keys that call COMMAND-SYM."
   (let (mode-lines
         global-lines)
-    (--each (helpful--keymaps-containing-aliases command-sym)
+    (--each (helpful--keymaps-containing-aliases command-sym aliases)
       (-let [(map . keys) it]
         (dolist (key keys)
           (push
@@ -1411,9 +1503,10 @@ along with its position.
 Moves point in BUF."
   (with-current-buffer buf
     (goto-char pos)
-    (let* ((ppss (syntax-ppss)))
-      (unless (zerop (syntax-ppss-depth ppss))
-        (beginning-of-defun)))
+    (let* ((ppss (syntax-ppss))
+           (outer-sexp-posns (nth 9 ppss)))
+      (when outer-sexp-posns
+        (goto-char (car outer-sexp-posns))))
     (list (point) (-take 2 (read buf)))))
 
 (defun helpful--count-values (items)
@@ -1428,7 +1521,7 @@ E.g. (x x y z y) -> ((x . 2) (y . 2) (z . 1))"
 (defun helpful--without-advice (sym)
   "Given advised function SYM, return the function object
 without the advice."
-  (advice--cdr
+  (advice--cd*r
    (advice--symbol-function sym)))
 
 (defun helpful--advised-p (sym)
@@ -1655,11 +1748,11 @@ OBJ may be a symbol or a compiled function object."
             "special form"
             'helpful-info-button
             'info-node "(elisp)Special Forms"))
-	  (keyboard-macro-button
-	   (helpful--button
-	    "keyboard macro"
-	    'helpful-info-button
-	    'info-node "(elisp)Keyboard Macros"))
+          (keyboard-macro-button
+           (helpful--button
+            "keyboard macro"
+            'helpful-info-button
+            'info-node "(elisp)Keyboard Macros"))
           (interactive-button
            (helpful--button
             "interactive"
@@ -1708,7 +1801,7 @@ OBJ may be a symbol or a compiled function object."
                       'callable-p callable-p)))
             ((not callable-p) "variable")
             ((macrop sym) "macro")
-	    ((helpful--kbd-macro-p sym) keyboard-macro-button)
+            ((helpful--kbd-macro-p sym) keyboard-macro-button)
             (t "function")))
           (defined
             (cond
@@ -1730,7 +1823,9 @@ OBJ may be a symbol or a compiled function object."
     (s-word-wrap
      70
      (format "%s is %s %s %s."
-             (if (symbolp sym) sym "This lambda")
+             (if (symbolp sym)
+                 (format "%S" sym)
+               "This lambda")
              description kind defined))))
 
 (defun helpful--callees (form)
@@ -1831,6 +1926,14 @@ may contain duplicates."
      (-flatten
       (-map #'helpful--callees-1 (cdr form)))))))
 
+(defun helpful--ensure-loaded ()
+  "Ensure the symbol associated with the current buffer has been loaded."
+  (when (and helpful--callable-p
+             (symbolp helpful--sym))
+    (let ((fn-obj (helpful--without-advice helpful--sym)))
+      (when (autoloadp fn-obj)
+        (autoload-do-load fn-obj)))))
+
 (defun helpful-update ()
   "Update the current *Helpful* buffer to the latest
 state of the current symbol."
@@ -1838,10 +1941,17 @@ state of the current symbol."
   (cl-assert (not (null helpful--sym)))
   (unless (buffer-live-p helpful--associated-buffer)
     (setq helpful--associated-buffer nil))
-  (-let* ((inhibit-read-only t)
+  (helpful--ensure-loaded)
+  (-let* ((val
+           ;; Look at the value before setting `inhibit-read-only', so
+           ;; users can see the correct value of that variable.
+           (unless helpful--callable-p
+             (helpful--sym-value helpful--sym helpful--associated-buffer)))
+          (inhibit-read-only t)
           (start-line (line-number-at-pos))
           (start-column (current-column))
           (primitive-p (helpful--primitive-p helpful--sym helpful--callable-p))
+          (canonical-sym (helpful--canonical-symbol helpful--sym helpful--callable-p))
           (look-for-src (or (not primitive-p)
                             find-function-C-source-directory))
           ((buf pos opened)
@@ -1854,14 +1964,20 @@ state of the current symbol."
                          (buffer-file-name buf)))
           (references (helpful--calculate-references
                        helpful--sym helpful--callable-p
-                       source-path)))
+                       source-path))
+          (aliases (helpful--aliases helpful--sym helpful--callable-p)))
 
     (erase-buffer)
 
     (insert (helpful--summary helpful--sym helpful--callable-p buf pos))
 
+    (when (helpful--obsolete-info helpful--sym helpful--callable-p)
+      (insert
+       "\n\n"
+       (helpful--format-obsolete-info helpful--sym helpful--callable-p)))
+
     (when (and helpful--callable-p
-	       (not (helpful--kbd-macro-p helpful--sym)))
+               (not (helpful--kbd-macro-p helpful--sym)))
       (helpful--insert-section-break)
       (insert
        (helpful--heading "Signature")
@@ -1870,7 +1986,6 @@ state of the current symbol."
     (when (not helpful--callable-p)
       (helpful--insert-section-break)
       (let* ((sym helpful--sym)
-             (val (helpful--sym-value sym helpful--associated-buffer))
              (multiple-views-p
               (or (stringp val)
                   (keymapp val)
@@ -1950,7 +2065,7 @@ state of the current symbol."
       (helpful--insert-section-break)
       (insert
        (helpful--heading "Key Bindings")
-       (helpful--format-keys helpful--sym)))
+       (helpful--format-keys helpful--sym aliases)))
 
     (helpful--insert-section-break)
 
@@ -2034,18 +2149,20 @@ state of the current symbol."
           (insert " "))
         (insert (helpful--make-forget-button helpful--sym helpful--callable-p))))
 
-    (let ((aliases (helpful--aliases helpful--sym helpful--callable-p)))
-      (when aliases
-        (helpful--insert-section-break)
-        (insert
-         (helpful--heading "Aliases")
-         (s-join "\n" (--map (helpful--format-alias it helpful--callable-p)
-                             aliases)))))
+    (when aliases
+      (helpful--insert-section-break)
+      (insert
+       (helpful--heading "Aliases")
+       (s-join "\n" (--map (helpful--format-alias it helpful--callable-p)
+                           aliases))))
 
     (helpful--insert-section-break)
 
     (insert
-     (helpful--heading "Source Code")
+     (helpful--heading
+      (if (eq helpful--sym canonical-sym)
+          "Source Code"
+        "Alias Source Code"))
      (cond
       (source-path
        (concat
@@ -2071,7 +2188,11 @@ state of the current symbol."
       (insert
        (cond
         ((stringp source)
-         (helpful--syntax-highlight source (if primitive-p 'c-mode)))
+         (let ((mode (when primitive-p
+                       (pcase (file-name-extension source-path)
+                         ("c" 'c-mode)
+                         ("rs" (when (fboundp 'rust-mode) 'rust-mode))))))
+           (helpful--syntax-highlight source mode)))
         ((and (consp source) (eq (car source) 'closure))
          (helpful--syntax-highlight
           (concat ";; Closure converted to defun by helpful.\n"
@@ -2080,7 +2201,9 @@ state of the current symbol."
         (t
          (helpful--syntax-highlight
           (concat
-           ";; Source file is unknown, showing raw function object.\n"
+           (if (eq helpful--sym canonical-sym)
+               ";; Could not find source code, showing raw function object.\n"
+             ";; Could not find alias source code, showing raw function object.\n")
            (helpful--pretty-print source)))))))
 
     (helpful--insert-section-break)
@@ -2141,11 +2264,11 @@ For example, \"(some-func FOO &optional BAR)\"."
                       (s-join " " formatted-args)))
              ;; If it has multiple arguments, join them with spaces.
              (formatted-args
-              (format "(%s %s)" sym
+              (format "(%S %s)" sym
                       (s-join " " formatted-args)))
              ;; Otherwise, this function takes no arguments when called.
              (t
-              (format "(%s)" sym)))))
+              (format "(%S)" sym)))))
 
     ;; If the docstring ends with (fn FOO BAR), extract that.
     (-when-let (docstring (documentation sym))
@@ -2163,6 +2286,19 @@ For example, \"(some-func FOO &optional BAR)\"."
       ;; Otherwise, just use the signature from the source code.
       source-sig))))
 
+(defun helpful--format-obsolete-info (sym callable-p)
+  (-let [(use _ date) (helpful--obsolete-info sym callable-p)]
+    (helpful--propertize-quoted
+     (s-word-wrap
+      70
+      (format "This %s is obsolete%s%s"
+              (helpful--kind-name sym callable-p)
+              (if date (format " since %s" date)
+                "")
+              (cond ((stringp use) (concat "; " use))
+                    (use (format "; use `%s' instead." use))
+                    (t ".")))))))
+
 (defun helpful--docstring (sym callable-p)
   "Get the docstring for SYM.
 Note that this returns the raw docstring, including \\=\\=
@@ -2196,9 +2332,9 @@ escapes that are used by `substitute-command-keys'."
              (rx ": " eos)
              (format " (default: %s): " default-val)
              prompt)))
-    (read (completing-read prompt obarray
-                           predicate t nil nil
-                           default-val))))
+    (intern (completing-read prompt obarray
+                             predicate t nil nil
+                             default-val))))
 
 ;;;###autoload
 (defun helpful-function (symbol)
@@ -2268,6 +2404,20 @@ or :foo."
   (or (fboundp symbol)
       (helpful--variable-p symbol)))
 
+(defun helpful--convert-c-name (symbol var)
+  "Convert SYMBOL from a C name to an Elisp name.
+E.g. convert `Fmake_string' to `make-string' or
+`Vgc_cons_percentage' to `gc-cons-percentage'. Interpret
+SYMBOL as variable name if VAR, else a function name. Return
+nil if SYMBOL doesn't begin with \"F\" or \"V\"."
+  (let ((string (symbol-name symbol))
+        (prefix (if var "V" "F")))
+    (when (s-starts-with-p prefix string)
+      (intern
+       (s-chop-prefix
+        prefix
+        (s-replace "_" "-" string))))))
+
 ;;;###autoload
 (defun helpful-symbol (symbol)
   "Show help for SYMBOL, a variable, function or macro.
@@ -2286,6 +2436,14 @@ See also `helpful-callable' and `helpful-variable'."
     (helpful-callable symbol))
    ((boundp symbol)
     (helpful-variable symbol))
+   ((let ((fsym (helpful--convert-c-name symbol nil)))
+      (when (fboundp fsym)
+        (helpful-callable fsym)
+        fsym)))
+   ((let ((vsym (helpful--convert-c-name symbol t)))
+      (when (boundp vsym)
+        (helpful-variable vsym)
+        vsym)))
    (t
     (user-error "Not bound: %S" symbol))))
 
@@ -2326,7 +2484,7 @@ imenu."
   "Temporarily highlight region from START to END."
   (let ((overlay (make-overlay start end)))
     (overlay-put overlay 'face 'highlight)
-    (run-with-timer 1.0 nil 'delete-overlay overlay)))
+    (run-with-timer 1.5 nil 'delete-overlay overlay)))
 
 (defun helpful-visit-reference ()
   "Go to the reference at point."
