@@ -13,6 +13,14 @@
 
   (setq jka-compr-verbose nil))
 
+(defvar helpful-test-var nil)
+(define-obsolete-variable-alias 'helpful-test-var-obsolete
+  'helpful-test-var "23.1")
+
+(defun helpful-test-fun-obsolete ()
+  (declare (obsolete helpful "1.2.3"))
+  nil)
+
 (defun test-foo ()
   "Docstring here."
   nil)
@@ -133,7 +141,21 @@ symbol (not a form)."
   ;; `rename-buffer' is primitive, but it's advised by uniquify.
   (should (helpful--primitive-p 'rename-buffer t)))
 
+(ert-deftest helpful--without-advice ()
+  "Ensure we remove advice to get the underlying function."
+  ;; Removing the advice on an unadvised function should give us the
+  ;; same function.
+  (should
+   (eq (helpful--without-advice #'test-foo)
+       (indirect-function #'test-foo)))
+  ;; Removing the advice should give us an unadvised function.
+  (should
+   (not (helpful--advised-p
+         (helpful--without-advice #'test-foo-advised)))))
+
 (ert-deftest helpful-callable ()
+  ;; Functions. Also a regression test for #170.
+  (helpful-callable 'face-attribute)
   ;; We should not crash when looking at macros.
   (helpful-callable 'when)
   ;; Special forms should work too.
@@ -152,7 +174,7 @@ symbol (not a form)."
 (ert-deftest helpful--no-symbol-properties ()
   "Helpful should handle functions without any symbol properties."
   ;; Interactively evaluating this file will set edebug properties on
-  ;; test-foo, so remove all properties.
+  ;; test-foo-no-properties, so remove all properties.
   (setplist #'test-foo-no-properties nil)
 
   ;; This shouldn't throw any errors.
@@ -320,6 +342,8 @@ variables defined without `defvar'."
   (require 'python)
   ;; This test will fail if you already have python.el.gz open in your
   ;; Emacs instance.
+  (skip-unless (null (get-buffer "python.el.gz")))
+
   (-let [(buf pos opened) (helpful--definition 'python-indent-offset nil)]
     (should (bufferp buf))
     (should opened)))
@@ -340,6 +364,13 @@ variables defined without `defvar'."
         (-let [(buf pos opened) (helpful--definition 'test-foo-edebug-defn t)]
           (should buf))))))
 
+(ert-deftest helpful--definition-defstruct ()
+  "Ensure we find the position of struct functions."
+  (-let [(buf pos _)
+         (helpful--definition #'make-ert-test t)]
+    (should buf)
+    (should pos)))
+
 (ert-deftest helpful-variable ()
   "Smoke test for `helpful-variable'."
   (helpful-variable 'tab-width))
@@ -359,6 +390,12 @@ variables defined without `defvar'."
    (equal (helpful--signature 'some-unused-function)
           "(some-unused-function [Arg list not available until function definition is loaded.])")))
 
+(ert-deftest helpful--signature-space ()
+  "Ensure that symbols with spaces are handled correctly."
+  (should
+   (equal (helpful--signature 'helpful-test-fn-with\ space)
+          "(helpful-test-fn-with\\ space)")))
+
 (ert-deftest helpful--signature--advertised ()
   "Ensure that we respect functions that declare `advertised-calling-convention'."
   (should
@@ -455,6 +492,20 @@ associated a lambda with a keybinding."
     (with-current-buffer buf
       (helpful-update))))
 
+(ert-deftest helpful--obsolete-variable ()
+  "Test display of obsolete variable."
+  (let* ((var 'helpful-test-var-obsolete)
+         (info (helpful--format-obsolete-info var nil)))
+    (should
+     (equal info "This variable is obsolete since 23.1; use helpful-test-var instead."))))
+
+(ert-deftest helpful--obsolete-function ()
+  "Test display of obsolete function."
+  (let* ((fun 'helpful-test-fun-obsolete)
+         (info (helpful--format-obsolete-info fun t)))
+    (should
+     (equal info "This function is obsolete since 1.2.3; use helpful instead."))))
+
 (ert-deftest helpful--keymap-keys--sparse ()
   (let* ((parent-keymap (make-sparse-keymap))
          (keymap (make-sparse-keymap)))
@@ -588,7 +639,7 @@ in."
   (global-set-key (kbd "C-c M-S-d") #'helpful--dummy-command-alias)
 
   (unwind-protect
-      (let* ((keymaps (helpful--keymaps-containing-aliases #'helpful--dummy-command))
+      (let* ((keymaps (helpful--keymaps-containing-aliases #'helpful--dummy-command (helpful--aliases 'helpful--dummy-command t)))
              (global-keybindings (cdr (assoc "global-map" keymaps))))
         (should
          (equal global-keybindings (list "C-c M-S-c" "C-c M-S-d"))))
@@ -679,6 +730,17 @@ find the source code."
     (should
      (s-starts-with-p "if is a special form defined in" summary))))
 
+(defun helpful-test-fn-with\ space ()
+  42)
+
+(ert-deftest helpful--summary--symbol-with-space ()
+  "Ensure we correctly format symbols containing spaces."
+  (let* ((summary (helpful--summary 'helpful-test-fn-with\ space t nil nil)))
+    ;; Strip properties to make assertion messages more readable.
+    (set-text-properties 0 (1- (length summary)) nil summary)
+    (should
+     (s-starts-with-p "helpful-test-fn-with\\ space is a function" summary))))
+
 (ert-deftest helpful--bound-p ()
   ;; Functions.
   (should (helpful--bound-p 'message))
@@ -799,3 +861,33 @@ find the source code."
     (should (helpful--autoloaded-p 'rx-to-string buf))
     (when opened
       (kill-buffer buf))))
+
+(ert-deftest helpful--inhibit-read-only ()
+  (helpful-variable 'inhibit-read-only)
+  (should
+   (s-contains-p "Value\nnil" (buffer-string))))
+
+(ert-deftest helpful--convert-c-name ()
+  (should
+   (equal
+    'make-string
+    (helpful--convert-c-name 'Fmake_string nil)))
+  (should
+   (equal
+    'gc-cons-percentage
+    (helpful--convert-c-name 'Vgc_cons_percentage t)))
+  (should-not
+   (helpful--convert-c-name 'Fmake_string t))
+  (should-not
+   (helpful--convert-c-name 'Vgc_cons_percentage nil)))
+
+(ert-deftest helpful--loads-autoload-symbol ()
+  "When asked to describe an autoloaded symbol, just load it."
+  ;; This test assumes that you haven't loaded tetris.el.gz in your
+  ;; current instance.
+  (skip-unless (autoloadp (symbol-function 'tetris)))
+  ;; This is a regression test: `tetris' has `tetris-mode-map' in its
+  ;; docstring, so we can't display the mode map unless tetris.el.gz is
+  ;; loaded.
+  ;;
+  (helpful-function #'tetris))
